#
# Copyright (c) 2022 Arm Limited
# Copyright (c) 2022 Hanno Becker
# Copyright (c) 2023 Amin Abdulrahman, Matthias Kannwischer
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Hanno Becker <hannobecker@posteo.de>
#

"""SLOTHY heuristics

The one-shot SLOTHY approach tends to become computationally infeasible above
200 assembly instructions. To optimize kernels beyond that threshold, this
module provides heuristics to split the optimization problem into several
smaller-sizes problems amenable to one-shot SLOTHY.
"""

import math

from slothy.core.dataflow import DataFlowGraph as DFG
from slothy.core.dataflow import Config as DFGConfig, ComputationNode
from slothy.core.core import SlothyBase, Result, SlothyException
from slothy.helper import Permutation, SourceLine
from slothy.helper import binary_search, BinarySearchLimitException


class Heuristics:
    """Break down large optimization problems into smaller ones.

    The one-shot SLOTHY approach tends to become computationally infeasible above
    200 assembly instructions. To optimize kernels beyond that threshold, this
    class provides heuristics to split the optimization problem into several
    smaller-sizes problems amenable to one-shot SLOTHY.
    """

    @staticmethod
    def _optimize_binsearch_core(source, logger, conf, **kwargs):

        logger_name = logger.name.replace(".", "_")
        last_successful = None

        def try_with_stalls(stalls, timeout=None):
            nonlocal last_successful

            logger.info(f"Attempt optimization with max {stalls} stalls...")
            c = conf.copy()
            c.constraints.stalls_allowed = stalls

            if c.hints.ext_bsearch_remember_successes:
                c.hints.rename_hint_orig_rename = True
                c.hints.order_hint_orig_order = True

            if timeout is not None:
                c.timeout = timeout
            core = SlothyBase(conf.arch, conf.target, logger=logger, config=c)

            if last_successful is not None:
                src = last_successful
            else:
                src = source
            success = core.optimize(src, **kwargs)

            if success and c.hints.ext_bsearch_remember_successes:
                last_successful = core.result.code

            return success, core

        try:
            return binary_search(
                try_with_stalls,
                minimum=conf.constraints.stalls_minimum_attempt - 1,
                start=conf.constraints.stalls_first_attempt,
                threshold=conf.constraints.stalls_maximum_attempt,
                precision=conf.constraints.stalls_precision,
                timeout_below_precision=conf.constraints.stalls_timeout_below_precision,
            )

        except BinarySearchLimitException:
            logger.error("Exceeded stall limit without finding a working solution")
            logger.error("Here's what you asked me to optimize:")

            Heuristics._dump(
                "Original source code",
                source,
                logger=logger,
                err=True,
                no_comments=True,
            )
            logger.error("Configuration:")
            conf.log(logger.error)

            err_file = conf.log_dir + f"/{logger_name}_ERROR.s"
            with open(err_file, "w", encoding="utf-8") as f:
                conf.log(lambda line: f.write("// " + line + "\n"))
                f.write("\n".join(source))

            logger.error(f"Stored this information in {err_file}")

    @staticmethod
    def optimize_binsearch(source: list, logger: any, conf: any, **kwargs: any) -> any:
        """Optimize for minimum number of stalls, and potentially a secondary objective.

        The `variable_size` configuration option determines whether the minimiation of
        stalls happens internally or externally. Internal minimization means that the
        number of stalls is part of the model, and its minimization registered as the
        objective to the underlying solver. External minimization means that the number
        of stalls is statically fixed per one-shot SLOTHY optimization, and that an
        external binary search is used to minimize it.

        :param source: The source code to be optimized. Must be a list of
            SourceLine instances.
        :type source: list
        :param logger: The logger to be used
        :type logger: any
        :param conf: The configuration to apply. This fixed for all one-shot SLOTHY
            runs invoked by this call, except for the variation of the stall count.
        :type conf: any
        :param **kwargs: An optional list of parameters to the core optimize routine
        :type **kwargs: any
        :returns: The Result object for the succeeding optimization with the smallest
                  number of stalls.
        :rtype: any
        """
        flexible = not conf.constraints.functional_only

        if conf.variable_size:
            return Heuristics.optimize_binsearch_internal(
                source, logger, conf, flexible=flexible, **kwargs
            )

        return Heuristics.optimize_binsearch_external(
            source, logger, conf, flexible=flexible, **kwargs
        )

    @staticmethod
    def _log_reoptimization_failure(log):
        log.warning(
            "Re-optimization with objective at minimum number of stalls failed. "
            "By the non-deterministic nature of the optimization, this can happen. "
            "Will just pick previous result..."
        )

    @staticmethod
    def _log_input_output_warning(log):
        log.warning(
            "You are using SW pipelining without setting inputs_are_outputs=True."
            "This means that the last iteration of the loop may overwrite inputs "
            "to the loop (such as address registers), unless they are marked as "
            "reserved registers. If this is intended, ignore this warning. "
            "Otherwise, consider setting inputs_are_outputs=True to ensure that "
            "nothing that is used as an input to the loop is overwritten, "
            "not even in the last iteration."
        )

    @staticmethod
    def optimize_binsearch_external(
        source: list, logger: any, conf: any, flexible: bool = True, **kwargs: any
    ) -> any:
        """Externally optimize for minimum number of stalls, and potentially a secondary
        objective.

        This function uses an external binary search to find the minimum number of stalls
        for which a one-shot SLOTHY optimization succeeds. If the provided configuration
        has a secondary objective, it then re-optimizes the result for that secondary
        objective, fixing the minimal number of stalls.


        :param source: The source code to be optimized. Must be a list of SourceLine
            instances.
        :type source: list
        :param logger: The logger to be used.
        :type logger: any
        :param conf: The configuration to apply. This is fixed for all one-shot SLOTHY
            runs invoked by this call, except for variation of stall count.
        :type conf: any
        :param flexible: Indicates whether the number of stalls should be minimized
            through a binary search, or whether a single one-shot SLOTHY optimization
            for a fixed number of stalls (encoded in the configuration) should be
            conducted.
        :type flexible: bool
        :param **kwargs: An optional list of parameters to the core optimize routine
        :type **kwargs: any
        :return: A Result object representing the final optimization result.
        :rtype: any
        :raises SlothyException: If optimization fails.
        """

        if not flexible:
            core = SlothyBase(conf.arch, conf.target, logger=logger, config=conf)
            if not core.optimize(source):
                raise SlothyException("Optimization failed")
            return core.result

        logger.info("Perform external binary search for minimal number of stalls...")

        c = conf.copy()
        c.ignore_objective = True
        min_stalls, core = Heuristics._optimize_binsearch_core(
            source, logger, c, **kwargs
        )

        if conf.has_objective is False:
            return core.result

        logger.info(
            "Optimize again with minimal number of %d stalls, with objective...",
            min_stalls,
        )
        first_result = core.result

        core.config.ignore_objective = False
        success = core.retry()

        if not success:
            Heuristics._log_reoptimization_failure(logger)
            return first_result

        return core.result

    @staticmethod
    def optimize_binsearch_internal(
        source: list, logger: any, conf: any, flexible: bool = True, **kwargs: any
    ) -> any:
        """Internally optimize for minimum number of stalls, and potentially a secondary
        objective.

        This finds the minimum number of stalls for which a one-shot SLOTHY optimization
        succeeds.
        If the provided configuration has a secondary objective, it then re-optimizes the
        result for that secondary objective, fixing the minimal number of stalls.


        :param source: The source code to be optimized. Must be a list of SourceLine
            instances.
        :type source: list
        :param  logger: The logger to be used.
        :type logger: any
        :param conf: The configuration to apply. This is fixed for all one-shot SLOTHY
            runs invoked by this call, except for variation of stall count.
        :type conf: any
        :param flexible: Indicates whether the number of stalls should be minimized
            through a binary search, or whether a single one-shot SLOTHY optimization
            for a fixed number of stalls (encoded in the configuration) should be
            conducted.
        :type flexible: bool
        :param **kwargs: An optional list of parameters to the core optimize routine
        :type **kwargs: any
        :return: A Result object representing the final optimization result.
        :rtype: any
        : raises SlothyException: If optimization fails.
        """

        if not flexible:
            core = SlothyBase(conf.arch, conf.target, logger=logger, config=conf)
            if not core.optimize(source):
                raise SlothyException("Optimization failed")
            return core.result

        logger.info("Perform internal binary search for minimal number of stalls...")

        start_attempt = conf.constraints.stalls_first_attempt
        cur_attempt = start_attempt

        while True:
            c = conf.copy()
            c.variable_size = True
            c.constraints.stalls_allowed = cur_attempt

            logger.info("Attempt optimization with max %d stalls...", cur_attempt)

            core = SlothyBase(c.arch, c.target, logger=logger, config=c)
            success = core.optimize(source, **kwargs)

            if success:
                min_stalls = core.result.stalls
                break

            cur_attempt = max(1, cur_attempt * 2)
            if cur_attempt > conf.constraints.stalls_maximum_attempt:
                logger.error("Exceeded stall limit without finding a working solution")
                raise SlothyException("No solution found")

        logger.info(f"Minimum number of stalls: {min_stalls}")

        # Spill minimization is integrated into the stall minimization objective
        if conf.has_objective is False or conf.constraints.minimize_spills is True:
            return core.result

        logger.info(
            "Optimize again with minimal number of %d stalls, with objective...",
            min_stalls,
        )
        first_result = core.result

        success = core.retry(fix_stalls=min_stalls)
        if not success:
            Heuristics._log_reoptimization_failure(logger)
            return first_result

        return core.result

    @staticmethod
    def periodic(body: list, logger: any, conf: any) -> any:
        """Entrypoint for optimization of loops.

        If software pipelining is disabled, this function forwards to
        the straightline optimization via Heuristics.linear().

        If software pipelining is enabled but the halving heuristic
        is disabled, this function performs a one-shot SLOTHY optimization
        without heuristics.

        If software pipelining is enabled and the halving heuristic is
        enabled, this function optimizes the loop body via straightline
        optimization first, splits result as `[A;B]`, and optimizes
        `[B;A]` again via straightline optimizations. The optimized loop
        is then given by the preamble `A`, kernel `opt(B;A)`, and postamble
        `B`. The straightline optimizations applied in this heuristics are
        done via Heuristics.linear() and thus themselves subject to the
        splitting heuristic, if enabled.


        :param body: The loop body to be optimized. This must be a list of
            SourceLine instances.
        :type body: list
        :param logger: The logger to be used.
        :type logger: any
        :param conf: The configuration to be applied.
        :type conf: any

        :return: Tuple (preamble, kernel, postamble, num_exceptional_iterations)
            of preamble, kernel and postamble (each as a list of SourceLine
            objects), plus the number of iterations jointly accounted for by
            the preamble and postamble (the caller will need this to adjust the
            loop counter).
        :rtype: any
        """

        if conf.sw_pipelining.enabled and not conf.inputs_are_outputs:
            Heuristics._log_input_output_warning(logger)

        if conf.sw_pipelining.enabled:
            body = body * conf.sw_pipelining.unroll

        if conf.inputs_are_outputs:
            dfg = DFG(
                body, logger.getChild("dfg_generate_outputs"), DFGConfig(conf.copy())
            )
            conf.outputs = dfg.outputs
            conf.inputs_are_outputs = False

        # If we're not asked to do software pipelining, just forward to
        # the heuristics for linear optimization.
        if not conf.sw_pipelining.enabled:
            res = Heuristics.linear(body, logger=logger, conf=conf)
            return [], res.code, [], 0

        if conf.sw_pipelining.halving_heuristic:
            return Heuristics._periodic_halving(body, logger, conf)

        # 'Normal' software pipelining
        #
        # We first perform the core periodic optimization of the loop kernel,
        # and then separate passes for the optimization for the preamble and postamble

        # First step: Optimize loop kernel

        logger.debug("Optimize loop kernel...")
        c = conf.copy()
        c.inputs_are_outputs = True
        result = Heuristics.optimize_binsearch(body, logger.getChild("slothy"), c)

        conf.outputs = list(
            map(lambda o: result.output_renamings.get(o, o), conf.outputs)
        )

        num_exceptional_iterations = result.num_exceptional_iterations
        kernel = result.code
        assert SourceLine.is_source(kernel)

        # Second step: Separately optimize preamble and postamble

        preamble = result.preamble
        if conf.sw_pipelining.optimize_preamble:
            logger.debug("Optimize preamble...")
            Heuristics._dump("Preamble", preamble, logger)
            logger.debug("Dependencies within kernel: %s", result.kernel_input_output)
            c = conf.copy()
            c.outputs = result.kernel_input_output
            c.sw_pipelining.enabled = False
            res_preamble = Heuristics.linear(
                preamble, conf=c, logger=logger.getChild("preamble")
            )
            preamble = res_preamble.code

        postamble = result.postamble
        if conf.sw_pipelining.optimize_postamble:
            logger.debug("Optimize postamble...")
            Heuristics._dump("Preamble", postamble, logger)
            c = conf.copy()
            c.sw_pipelining.enabled = False
            res_postamble = Heuristics.linear(
                postamble, conf=c, logger=logger.getChild("postamble")
            )
            postamble = res_postamble.code

        return preamble, kernel, postamble, num_exceptional_iterations

    @staticmethod
    def linear(body: list, logger: any, conf: any) -> any:
        """Entrypoint for straightline optimization.

        If the split heuristic is disabled, this forwards to a one-shot optimization.

        If the split heuristic is enabled (conf.split_heuristic == True), the assembly
        input is optimized by successively applying one-shot optimizations to a
        'sliding window' of code.

        :param body: The assembly input to be optimized. This must be a list of
            SourceLine objects.
        :type body: list
        :param logger: The logger to be used.
        :type logger: any
        :param conf: The configuration to be applied. Software pipelining must be
            disabled.
        :type conf: any
        :return: A Result object representing the final optimization result.
        :rtype: any
        :raises SlothyException: If software pipelining is enabled.
        """
        assert SourceLine.is_source(body)
        if conf.sw_pipelining.enabled:
            raise SlothyException(
                "Linear heuristic should only be called " "with SW pipelining disabled"
            )

        Heuristics._dump("Starting linear optimization...", body, logger)

        # So far, we only implement one heuristic: The splitting heuristic --
        # If that's disabled, just forward to the core optimization
        if not conf.split_heuristic:
            return Heuristics.optimize_binsearch(body, logger.getChild("slothy"), conf)

        return Heuristics._split(body, logger, conf)

    @staticmethod
    def _naive_reordering(body, logger, conf, use_latency_depth=False):

        if use_latency_depth:
            depth_str = "latency depth"
        else:
            depth_str = "depth"

        logger.info(f"Perform naive interleaving by {depth_str}... ")
        old = body.copy()
        le = len(body)
        dfg = DFG(body, logger.getChild("dfg"), DFGConfig(conf.copy()), parsing_cb=True)
        insts = [dfg.nodes[i] for i in range(le)]

        if use_latency_depth is True:
            # Calculate latency-depth of instruction nodes
            nodes_by_depth = dfg.nodes.copy()
            nodes_by_depth.sort(key=lambda t: t.depth)
            for t in dfg.nodes_all:
                t.latency_depth = 0

            def get_latency(tp, t):
                if tp.src.is_virtual:
                    return 0
                return conf.target.get_latency(tp.src.inst, tp.idx, t.inst)

            for t in nodes_by_depth:
                srcs = t.src_in + t.src_in_out
                t.latency_depth = max(
                    map(
                        lambda tp, t=t: tp.src.latency_depth + get_latency(tp, t), srcs
                    ),
                    default=0,
                )

        def get_depth(t):
            if use_latency_depth is False:
                pre_depth = t.depth
            else:
                pre_depth = t.latency_depth
            scale = float(t.inst.source_line.tags.get("naive_interleaving_scale", 1.0))
            return int(pre_depth * scale)

        depths = [get_depth(dfg.nodes_by_id[i]) for i in range(le)]

        inputs = dfg.inputs.copy()
        outputs = conf.outputs.copy()

        perm = Permutation.permutation_id(le)

        def get_inputs(inst):
            return set(inst.args_in + inst.args_in_out)

        def get_outputs(inst):
            return set(inst.args_out + inst.args_in_out)

        joint_prev_inputs = {}
        joint_prev_outputs = {}

        strategy = conf.split_heuristic_preprocess_naive_interleaving_strategy

        def get_interleaving_class(j):
            return int(insts[j].inst.source_line.tags.get("interleaving_class", 0))

        if strategy == "alternate":
            # Compute target ratio between code classes
            sz_0 = max(
                len(list(filter(lambda j: get_interleaving_class(j) == 0, range(le)))),
                1,
            )
            sz_1 = max(
                len(list(filter(lambda j: get_interleaving_class(j) == 1, range(le)))),
                1,
            )
            target_ratio = sz_0 / sz_1

        for i in range(le):
            cur_joint_prev_inputs = set()
            cur_joint_prev_outputs = set()
            for j in range(i, le):
                joint_prev_inputs[j] = cur_joint_prev_inputs
                cur_joint_prev_inputs = cur_joint_prev_inputs.union(
                    get_inputs(insts[j].inst)
                )

                joint_prev_outputs[j] = cur_joint_prev_outputs
                cur_joint_prev_outputs = cur_joint_prev_outputs.union(
                    get_outputs(insts[j].inst)
                )

            # Find instructions which could, in principle, come next, without
            # any renaming
            def could_come_next(j):
                cur_outputs = get_outputs(insts[j].inst)
                prev_inputs = joint_prev_inputs[j]

                cur_inputs = get_inputs(insts[j].inst)
                prev_outputs = joint_prev_outputs[j]

                ok = (
                    len(cur_outputs.intersection(prev_inputs)) == 0
                    and len(cur_inputs.intersection(prev_outputs)) == 0
                )

                return ok

            candidate_idxs = list(filter(could_come_next, range(i, le)))
            logger.debug(f"Potential next candidates: {candidate_idxs}")

            def pick_candidate(candidate_idxs):

                if strategy == "depth":
                    candidate_depths = list(map(lambda j: depths[j], candidate_idxs))
                    logger.debug("Candidate %s: %s", depth_str, candidate_depths)
                    choice_idx = candidate_idxs[
                        candidate_depths.index(min(candidate_depths))
                    ]

                else:
                    assert strategy == "alternate"

                    sz_0 = max(
                        len(
                            list(
                                filter(
                                    lambda j: get_interleaving_class(j) == 0, range(i)
                                )
                            )
                        ),
                        1,
                    )
                    sz_1 = max(
                        len(
                            list(
                                filter(
                                    lambda j: get_interleaving_class(j) == 1, range(i)
                                )
                            )
                        ),
                        1,
                    )

                    candidates_0 = filter(
                        lambda j: get_interleaving_class(j) == 0, candidate_idxs
                    )
                    candidates_1 = filter(
                        lambda j: get_interleaving_class(j) == 1, candidate_idxs
                    )

                    current_ratio = sz_0 / sz_1

                    c0 = next(candidates_0, None)
                    c1 = next(candidates_1, None)

                    if current_ratio > target_ratio and c1 is not None:
                        choice_idx = c1
                    elif c0 is not None:
                        choice_idx = c0
                    else:
                        choice_idx = candidate_idxs[0]

                return choice_idx

            def move_entry_forward(lst, idx_from, idx_to):
                entry = lst[idx_from]
                del lst[idx_from]
                return lst[:idx_to] + [entry] + lst[idx_to:]

            choice_idx = None
            while choice_idx is None:
                choice_idx = pick_candidate(candidate_idxs)
                insts = move_entry_forward(insts, choice_idx, i)

            local_perm = Permutation.permutation_move_entry_forward(le, choice_idx, i)
            perm = Permutation.permutation_comp(local_perm, perm)

            body = list(map(ComputationNode.to_source_line, insts))
            depths = move_entry_forward(depths, choice_idx, i)
            body[i].set_length(100).set_comment(f"{depth_str} {depths[i]}")
            Heuristics._dump("New code", body, logger)

        # Selfcheck
        res = Result(conf)
        res.orig_code = old
        res.code = body.copy()
        res.codesize_with_bubbles = le
        res.success = True
        res.reordering_with_bubbles = perm
        res.input_renamings = {s: s for s in inputs}
        res.output_renamings = {s: s for s in outputs}
        res.valid = True
        res.selfcheck(logger.getChild("naive_interleaving_selfcheck"))

        res.offset_fixup(logger.getChild("naive_interleaving_fixup"))
        body = res.code_raw

        Heuristics._dump("Before naive interleaving", old, logger)
        Heuristics._dump("After naive interleaving", body, logger)
        return body, perm

    @staticmethod
    def _idxs_from_fractions(fraction_lst, body):
        return [round(f * len(body)) for f in fraction_lst]

    @staticmethod
    def _get_ssa_form(body, logger, conf):
        logger.info("Transform DFG into SSA...")
        dfg = DFG(
            body, logger.getChild("dfg_ssa"), DFGConfig(conf.copy()), parsing_cb=True
        )
        dfg.ssa()
        ssa = [ComputationNode.to_source_line(t) for t in dfg.nodes]
        return ssa

    @staticmethod
    def _split_inner(body, logger, conf, ssa=False):

        le = len(body)
        if le == 0:
            return body
        log = logger.getChild("split")

        # Allow to proceed in steps
        split_factor = conf.split_heuristic_factor

        orig_body = body.copy()

        if conf.split_heuristic_preprocess_naive_interleaving:

            if ssa:
                body = Heuristics._get_ssa_form(body, logger, conf)
                Heuristics._dump("Code in SSA form:", body, logger, err=True)

            body, perm = Heuristics._naive_reordering(
                body,
                log,
                conf,
                use_latency_depth=(
                    conf.split_heuristic_preprocess_naive_interleaving_by_latency
                ),
            )

            if ssa:
                log.debug("Remove symbolics after SSA...")
                c = conf.copy()
                c.constraints.allow_reordering = False
                c.constraints.functional_only = True
                body = SourceLine.reduce_source(body)
                result = Heuristics.optimize_binsearch(
                    body, log.getChild("remove_symbolics"), conf=c
                )
                body = result.code
                body = SourceLine.reduce_source(body)
        else:
            perm = Permutation.permutation_id(le)

        def print_intarr(arr, ll, vals=50):
            m = max(10, max(arr))
            start_idxs = [(ll * i) // vals for i in range(vals)]
            end_idxs = [(ll * (i + 1)) // vals for i in range(vals)]
            avgs = []
            for s, e in zip(start_idxs, end_idxs):
                if s == e:
                    continue
                avg = sum(arr[s:e]) // (e - s)
                avgs.append(avg)
                log.info(f"[{s:3d}-{e:3d}]: {'*'*avg}{'.'*(m-avg)} ({avg})")

        def print_stalls(stalls, le):
            chunk_len = int(le // split_factor)
            # Convert stalls into 01 valued function
            stalls_arr = [i in stalls for i in range(le)]
            for v in stalls_arr:
                assert v in {0, 1}
            stalls_cumulative = [
                sum(
                    stalls_arr[
                        max(0, i - math.floor(chunk_len / 2)) : i
                        + math.ceil(chunk_len / 2)
                    ]
                )
                for i in range(le)
            ]
            print_intarr(stalls_cumulative, le)

        def optimize_chunk(start_idx, end_idx, body, stalls, show_stalls=True):
            """Optimizes a sub-chunks of the given snippet, delimited by pairs
            of start and end indices provided as arguments. Input/output register
            names stay intact -- in particular, overlapping chunks are allowed."""

            cur_pre = body[:start_idx]
            cur_body = body[start_idx:end_idx]
            cur_post = body[end_idx:]

            if not conf.split_heuristic_optimize_seam:
                prefix_len = 0
                suffix_len = 0
            else:
                prefix_len = min(len(cur_pre), conf.split_heuristic_optimize_seam)
                suffix_len = min(len(cur_post), conf.split_heuristic_optimize_seam)
                cur_prefix = cur_pre[-prefix_len:] if prefix_len > 0 else []
                cur_suffix = cur_post[:suffix_len]
                cur_body = cur_prefix + cur_body + cur_suffix
                cur_pre = cur_pre[:-prefix_len] if prefix_len > 0 else cur_pre
                cur_post = cur_post[suffix_len:]

            pre_pad = len(cur_pre)
            post_pad = len(cur_post)

            Heuristics._dump(
                f"Optimizing chunk [{start_idx}-{prefix_len}:{end_idx}+{suffix_len}]",
                cur_body,
                log,
            )
            if prefix_len > 0:
                Heuristics._dump("Using prefix", cur_prefix, log)
            if suffix_len > 0:
                Heuristics._dump("Using suffix", cur_suffix, log)

            # Find dependencies of rest of body

            dfgc = DFGConfig(conf.copy())
            dfgc.outputs = set(dfgc.outputs).union(conf.outputs)
            cur_outputs = DFG(cur_post, log.getChild("dfg_infer_outputs"), dfgc).inputs

            c = conf.copy()
            c.rename_inputs = {"other": "static"}  # No renaming
            c.rename_outputs = {"other": "static"}  # No renaming
            c.inputs_are_outputs = False
            c.outputs = cur_outputs

            result = Heuristics.optimize_binsearch(
                cur_body,
                log.getChild(f"{start_idx}_{end_idx}"),
                c,
                prefix_len=prefix_len,
                suffix_len=suffix_len,
            )
            Heuristics._dump(f"New chunk [{start_idx}:{end_idx}]", result.code, log)
            new_body = cur_pre + SourceLine.reduce_source(result.code) + cur_post

            perm = Permutation.permutation_pad(result.reordering, pre_pad, post_pad)

            keep_stalls = {
                i
                for i in stalls
                if i < start_idx - prefix_len or i >= end_idx + suffix_len
            }
            new_stalls = keep_stalls.union(
                map(lambda i: i + start_idx - prefix_len, result.stall_positions)
            )

            if show_stalls:
                print_stalls(new_stalls, le)

            return new_body, new_stalls, len(result.stall_positions), perm

        def optimize_chunks_many(
            start_end_idx_lst,
            body,
            stalls,
            abort_stall_threshold_high=None,
            abort_stall_threshold_low=None,
            **kwargs,
        ):
            perm = Permutation.permutation_id(len(body))
            for start_idx, end_idx in start_end_idx_lst:
                body, stalls, cur_stalls, local_perm = optimize_chunk(
                    start_idx, end_idx, body, stalls, **kwargs
                )
                perm = Permutation.permutation_comp(local_perm, perm)
                if (
                    abort_stall_threshold_high is not None
                    and cur_stalls > abort_stall_threshold_high
                ):
                    break
                if (
                    abort_stall_threshold_low is not None
                    and cur_stalls < abort_stall_threshold_low
                ):
                    break
            return body, stalls, perm

        cur_body = body

        def make_idx_list_consecutive(factor, increment):
            chunk_len = 1 / factor
            cur_start = 0
            cur_end = 0
            start_pos = []
            end_pos = []
            while cur_end < 1.0:
                cur_end = cur_start + chunk_len
                cur_end = min(cur_end, 1.0)
                start_pos.append(cur_start)
                end_pos.append(cur_end)

                cur_start += increment

            def not_empty(x):
                return x[0] != x[1]

            idx_lst = zip(
                Heuristics._idxs_from_fractions(start_pos, cur_body),
                Heuristics._idxs_from_fractions(end_pos, cur_body),
            )
            idx_lst = list(filter(not_empty, idx_lst))
            return idx_lst

        stalls = set()
        increment = 1 / split_factor

        # First, do a 'dry run' solely for finding the initial 'stall map'
        if conf.split_heuristic_repeat > 0:
            orig_conf = conf.copy()
            conf.constraints.allow_reordering = False
            conf.constraints.allow_renaming = False
            idx_lst = make_idx_list_consecutive(split_factor, increment)
            cur_body, stalls, _ = optimize_chunks_many(
                idx_lst, cur_body, stalls, show_stalls=False
            )
            conf = orig_conf.copy()

            log.info("Initial stalls")
            print_stalls(stalls, le)

        if conf.split_heuristic_stepsize is None:
            increment = 1 / (2 * split_factor)
        else:
            increment = conf.split_heuristic_stepsize

        # Remember inputs and outputs
        dfgc = DFGConfig(conf.copy())
        outputs = conf.outputs.copy()
        inputs = DFG(orig_body, log.getChild("dfg_infer_inputs"), dfgc).inputs.copy()

        for _ in range(conf.split_heuristic_repeat):

            cur_body = SourceLine.reduce_source(cur_body)

            if conf.split_heuristic_chunks:
                start_pos = [x[0] for x in conf.split_heuristic_chunks]
                end_pos = [x[1] for x in conf.split_heuristic_chunks]
                idx_lst = zip(
                    Heuristics._idxs_from_fractions(start_pos, cur_body),
                    Heuristics._idxs_from_fractions(end_pos, cur_body),
                )

                def not_empty(x):
                    return x[0] != x[1]

                idx_lst = list(filter(not_empty, idx_lst))
            else:
                idx_lst = make_idx_list_consecutive(split_factor, increment)
                if conf.split_heuristic_bottom_to_top is True:
                    idx_lst.reverse()

            cur_body, stalls, local_perm = optimize_chunks_many(
                idx_lst,
                cur_body,
                stalls,
                abort_stall_threshold_high=conf.split_heuristic_abort_cycle_at_high,
                abort_stall_threshold_low=conf.split_heuristic_abort_cycle_at_low,
            )
            perm = Permutation.permutation_comp(local_perm, perm)

        # Check complete result
        res = Result(conf)
        res.orig_code = orig_body
        res.code = SourceLine.reduce_source(cur_body).copy()
        res.codesize_with_bubbles = res.codesize
        res.success = True
        res.reordering_with_bubbles = perm
        res.input_renamings = {s: s for s in inputs}
        res.output_renamings = {s: s for s in outputs}
        res.valid = True
        res.selfcheck(log.getChild("split_heuristic_full"))

        # Estimate performance of final code
        if conf.split_heuristic_estimate_performance:
            conf2 = conf.copy()
            conf2.constraints.allow_renaming = False
            conf2.constraints.allow_reordering = False
            conf2.variable_size = True
            stall_res = Heuristics.optimize_binsearch(
                res.code, logger.getChild("split_estimtate_perf"), conf2
            )
            if stall_res.success is False:
                log.error(
                    "Stall-estimate for final code after split heuristic failed"
                    " -- should not happen? Maybe increase timeout?"
                    " Just returning the result without stall-estimate."
                )
            else:
                res2 = Result(conf2)
                res2.orig_code = orig_body
                res2.code = res.code_raw
                res2.codesize_with_bubbles = stall_res.codesize_with_bubbles
                res2.success = True
                # Compose actual code reordering from split heuristic with
                # bubble-introducing (order-preserving) map
                res2.reordering_with_bubbles = {
                    i: stall_res.reordering_with_bubbles[res.reordering_with_bubbles[i]]
                    for i in range(res.codesize)
                }
                res2.input_renamings = {s: s for s in inputs}
                res2.output_renamings = {s: s for s in outputs}
                res2.valid = True
                res2.selfcheck(log.getChild("split_heuristic_full_with_stalls"))

                res = res2

        return res

    @staticmethod
    def _split(body, logger, conf):
        c = conf.copy()

        # Focus on the chosen subregion
        body = SourceLine.reduce_source(body)

        if c.split_heuristic_region == [0.0, 1.0]:
            return Heuristics._split_inner(body, logger, c)

        inputs = DFG(body, logger.getChild("dfg_generate_inputs"), DFGConfig(c)).inputs

        start_end_idxs = Heuristics._idxs_from_fractions(c.split_heuristic_region, body)
        start_idx = start_end_idxs[0]
        end_idx = start_end_idxs[1]

        pre = body[:start_idx]
        partial_body = body[start_idx:end_idx]
        post = body[end_idx:]

        # Adjust the outputs
        c.outputs = DFG(
            post, logger.getChild("dfg_generate_outputs"), DFGConfig(c)
        ).inputs
        c.inputs_are_outputs = False

        res = Heuristics._split_inner(partial_body, logger, c)
        new_partial_body = res.code

        pre_pad = len(pre)
        post_pad = len(post)
        perm = Permutation.permutation_pad(res.reordering, pre_pad, post_pad)

        new_body = SourceLine.reduce_source(pre + new_partial_body + post)

        res2 = Result(conf)
        res2.orig_code = body.copy()
        res2.code = new_body
        res2.codesize_with_bubbles = pre_pad + post_pad + res.codesize_with_bubbles
        res2.success = True
        res2.reordering_with_bubbles = perm
        res2.input_renamings = {s: s for s in inputs}
        res2.output_renamings = {s: s for s in conf.outputs}
        res2.valid = True
        res2.selfcheck(logger.getChild("split"))

        return res2

    @staticmethod
    def _dump(name, s, logger, err=False, no_comments=False):
        assert SourceLine.is_source(s)
        s = [line.to_string() for line in s]

        def strip_comments(sl):
            return [s.split("//")[0].strip() for s in sl]

        fun = logger.debug if not err else logger.error
        fun(f"Dump: {name} (size {len(s)})")
        if no_comments:
            s = strip_comments(s)
        for line in s:
            fun(f"> {line}")

    @staticmethod
    def _periodic_halving(body, logger, conf):

        assert conf is not None
        assert conf.sw_pipelining.enabled
        assert conf.sw_pipelining.halving_heuristic

        body = SourceLine.reduce_source(body)

        # Find kernel dependencies
        kernel_deps = DFG(
            body, logger.getChild("dfg_kernel_deps"), DFGConfig(conf.copy())
        ).inputs

        # First step: Optimize loop kernel, but without software pipelining
        c = conf.copy()
        c.sw_pipelining.enabled = False
        c.inputs_are_outputs = True
        c.outputs = c.outputs.union(kernel_deps)

        if not conf.sw_pipelining.halving_heuristic_split_only:
            res_halving_0 = Heuristics.linear(body, logger.getChild("slothy"), conf=c)

            # Split resulting kernel as [A;B] and synthesize result structure
            # as if SW pipelining has been used and the result would have been
            # [B;A], with preamble A and postamble B.
            #
            # Run the normal SW-pipelining selfcheck on this result.
            #
            # The overall goal here is to produce a result structure that's structurally
            # the same as for normal SW pipelining, including checks and visualization.
            #
            # TODO: The 2nd optimization step below does not yet produce a Result
            # structure.
            reordering = res_halving_0.reordering
            codesize = res_halving_0.codesize

            def rotate_pos(p):
                return p - (codesize // 2)

            def is_pre(i):
                return rotate_pos(reordering[i]) < 0

            kernel = SourceLine.reduce_source(res_halving_0.code)
            preamble = kernel[: codesize // 2]
            postamble = kernel[codesize // 2 :]

            # Swap halves around and consider new kernel [B;A]
            kernel = postamble + preamble

            dfgc = DFGConfig(c.copy())
            dfgc.inputs_are_outputs = False
            core_out = DFG(postamble, logger.getChild("dfg_kernel_deps"), dfgc).inputs

            dfgc = DFGConfig(conf.copy())
            dfgc.inputs_are_outputs = True
            dfgc.outputs = core_out
            new_kernel_deps = DFG(
                kernel, logger.getChild("dfg_kernel_deps"), dfgc
            ).inputs

            c2 = c.copy()
            c2.sw_pipelining.enabled = True

            reordering1 = {i: rotate_pos(reordering[i]) for i in range(codesize)}
            pre_core_post_dict1 = {
                i: (is_pre(i), not is_pre(i), False) for i in range(codesize)
            }

            res = Result(c2)
            res.orig_code = body
            res.code = kernel
            res.preamble = preamble
            res.postamble = postamble
            res.kernel_input_output = new_kernel_deps
            res.codesize_with_bubbles = res_halving_0.codesize_with_bubbles
            res.reordering_with_bubbles = reordering1
            res.pre_core_post_dict = pre_core_post_dict1
            res.input_renamings = {s: s for s in kernel_deps}
            res.output_renamings = {s: s for s in c.outputs}
            res.success = True
            res.valid = True

            # Check result as if it has been produced by SW pipelining run
            res.selfcheck(logger.getChild("halving_heuristic_1"))

        else:
            logger.info("Halving heuristic: Split-only -- no optimization")
            codesize = len(body)
            preamble = body[: codesize // 2]
            postamble = body[codesize // 2 :]
            kernel = postamble + preamble

            dfgc = DFGConfig(c.copy())
            dfgc.inputs_are_outputs = False
            kernel_deps = DFG(
                postamble, logger.getChild("dfg_kernel_deps"), dfgc
            ).inputs

            dfgc = DFGConfig(conf.copy())
            dfgc.inputs_are_outputs = True
            kernel_deps = DFG(kernel, logger.getChild("dfg_kernel_deps"), dfgc).inputs

        #
        # Second step:
        # Optimize the loop body _again_, but  swap the two loop halves to that
        # successive iterations can be interleaved somewhat.
        #
        # The benefit of this approach is that we never call SLOTHY with generic SW
        # pipelining, which is computationally significantly more complex than 'normal'
        # optimization. We do still enable SW pipelining in SLOTHY if
        # `halving_heuristic_periodic` is set, but this is only to make SLOTHY consider
        # the 'seam' between iterations -- since we unset `allow_pre/post`, SLOTHY does
        # not consider any loop interleaving.
        #

        # If the optimized loop body is [A;B], we now optimize [B;A], that is, the late
        # half of one iteration followed by the early half of the successive iteration.
        # The hope is that this enables good interleaving even without calling SLOTHY in
        # SW pipelining mode.

        logger.info(
            "Apply halving heuristic to optimize two halves "
            "of consecutive loop kernels..."
        )

        # The 'periodic' version considers the 'seam' between iterations; otherwise, we
        # consider [B;A] as a non-periodic snippet, which may still lead to stalls at the
        # loop boundary.

        if conf.sw_pipelining.halving_heuristic_periodic:
            c = conf.copy()
            c.inputs_are_outputs = True
            c.sw_pipelining.minimize_overlapping = False
            c.sw_pipelining.enabled = True  # SW pipelining enabled, but ...
            c.sw_pipelining.allow_pre = False  # - no early instructions
            c.sw_pipelining.allow_post = False  # - no late instructions
            # Just make sure to consider loop boundary
            kernel = Heuristics.optimize_binsearch(
                kernel, logger.getChild("periodic heuristic"), conf=c
            ).code
        elif not conf.sw_pipelining.halving_heuristic_split_only:
            c = conf.copy()
            c.outputs = new_kernel_deps
            c.inputs_are_outputs = True
            c.sw_pipelining.enabled = False

            res_halving_1 = Heuristics.linear(
                kernel, logger.getChild("heuristic"), conf=c
            )
            final_kernel = res_halving_1.code

            reordering2 = res_halving_1.reordering_with_bubbles

            c2 = conf.copy()

            def get_reordering2(i):
                is_pre = res.pre_core_post_dict[i][0]
                p = reordering2[res.periodic_reordering[i]]
                if is_pre:
                    p -= res_halving_1.codesize_with_bubbles
                return p

            reordering2 = {i: get_reordering2(i) for i in range(codesize)}

            res2 = Result(c2)
            res2.orig_code = body
            res2.code = SourceLine.reduce_source(final_kernel)
            res2.kernel_input_output = new_kernel_deps
            res2.codesize_with_bubbles = res_halving_1.codesize_with_bubbles
            res2.reordering_with_bubbles = reordering2
            res2.pre_core_post_dict = pre_core_post_dict1
            res2.input_renamings = res.input_renamings
            res2.output_renamings = res.output_renamings

            new_preamble = [
                final_kernel[i]
                for i in range(res2.codesize)
                if res2.is_pre(i, original_program_order=False) is True
            ]
            new_postamble = [
                final_kernel[i]
                for i in range(res2.codesize)
                if res2.is_pre(i, original_program_order=False) is False
            ]

            res2.preamble = new_preamble
            res2.postamble = new_postamble
            res2.success = True
            res2.valid = True

            # TODO: This does not yet work since there can be renaming at the boundary
            # between preamble and postamble that we don't account for in the selfcheck.
            # res2.selfcheck(logger.getChild("halving_heuristic_2"))

            kernel = final_kernel

        num_exceptional_iterations = 1
        return preamble, kernel, postamble, num_exceptional_iterations



#
# Copyright (c) 2022 Arm Limited
# Copyright (c) 2022 Hanno Becker
# Copyright (c) 2023 Amin Abdulrahman, Matthias Kannwischer
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Hanno Becker <hannobecker@posteo.de>
#

#
# Experimental model for Apple M1 efficiency "Icestorm" CPU cores
# Based on data by https://dougallj.github.io/applecpu/icestorm.html
#

from enum import Enum
from itertools import combinations, product
from slothy.helper import lookup_multidict

from slothy.targets.aarch64.aarch64_neon import (
    find_class,
    Ldr_X,
    Str_X,
    Ldr_Q,
    Str_Q,
    vuzp1,
    vuzp2,
    vzip1,
    vmov,
    vand,
    vadd,
    VShiftImmediateBasic,
    vusra,
    vmul,
    vumlal,
    vumull,
    vmla,
    vmla_lane,
    vmls,
    vmls_lane,
    vqrdmulh,
    vqrdmulh_lane,
    vqdmulh_lane,
    ASimdCompare,
    umov_d,
    mov_d,
    mov_d01,
    mov_b00,
    fcsel,
    St4,
    Ld4,
    St3,
    Ld3,
    q_ldr1_stack,
    Q_Ld2_Lane_Post_Inc,
    Mov_xtov_d,
    d_stp_stack_with_inc,
    d_str_stack_with_inc,
    b_ldr_stack_with_inc,
    d_ldr_stack_with_inc,
    is_qform_form_of,
    is_dform_form_of,
    add,
    add_imm,
    add_shifted,
    add2,
    umull_wform,
    mul_wform,
    umaddl_wform,
    lsr,
    bic,
    bic_reg,
    add_sp_imm,
    and_imm,
    movk_imm,
    sub,
    mov,
    asr_wform,
    and_imm_wform,
    lsr_wform,
    eor_wform,
    bfi,
    nop,
    tst_wform,
    subs_wform,
    x_stp_with_imm_sp,
    w_stp_with_imm_sp,
    ldr_const,
    ldr_sxtw_wform,
    vmul_lane,
    vbic,
    VShiftImmediateRounding,
    vsub,
    trn1,
    trn2,
    Vins,
    vzip2,
    x_ldr_stack_imm,
    x_str_sp_imm,
    ubfx,
)

# 4 uops per cycle
# Most arithmetic instructions consist of 1 uop, SIMD loads/stores are made up of multiple
issue_rate = 4


class ExecutionUnit(Enum):
    SCALAR_I0 = (0,)
    SCALAR_I1 = (1,)
    SCALAR_I2 = (2,)
    SCALAR_M = (2,)
    LSU0 = (3,)
    LU0 = (4,)
    VEC0 = (5,)
    VEC1 = (6,)

    def __repr__(self):
        return self.name

    def I():  # noqa: E743
        return [
            ExecutionUnit.SCALAR_I0,
            ExecutionUnit.SCALAR_I1,
            ExecutionUnit.SCALAR_I2,
        ]

    def M():
        return [ExecutionUnit.SCALAR_M]

    def V():
        return [ExecutionUnit.VEC0, ExecutionUnit.VEC1]

    def LSU():
        return [ExecutionUnit.LSU0]

    def LOAD():
        return [ExecutionUnit.LSU0, ExecutionUnit.LU0]

    def STORE():
        return [ExecutionUnit.LSU0]


#  Opaque functions called by SLOTHY to add further microarchitecture-
# specific constraints which are not encapsulated by the general framework.


def add_further_constraints(slothy):
    _ = slothy


def has_min_max_objective(config):
    return False


def get_min_max_objective(slothy):
    return


# TODO: Copy-pasted from A72 model and roughly adjusted


execution_units = {
    # Neon Arithmetic
    (
        vmul,
        vmul_lane,
        vmla,
        vmla_lane,
        vmls,
        vmls_lane,
        vqrdmulh,
        vqrdmulh_lane,
        vqdmulh_lane,
        vumull,
        vumlal,
        vusra,
        vand,
        vbic,
        ASimdCompare,
        VShiftImmediateBasic,
        VShiftImmediateRounding,
    ): ExecutionUnit.V(),
    (vadd, vsub, trn1, trn2): ExecutionUnit.V(),
    Vins: ExecutionUnit.V(),  # guessed
    (umov_d, mov_d): ExecutionUnit.V(),  # guessed
    (mov_d01, mov_b00): ExecutionUnit.V(),  # guessed
    fcsel: ExecutionUnit.V(),
    # Neon Load/Store
    St4: list(map(list, product(ExecutionUnit.STORE(), ExecutionUnit.V()))),  # guessed
    Ld4: [
        list(ll[0] + (ll[1],))
        for ll in map(
            list, (product(combinations(ExecutionUnit.LOAD(), 2), ExecutionUnit.V()))
        )
    ],  # guessed
    St3: list(map(list, product(ExecutionUnit.STORE(), ExecutionUnit.V()))),  # guessed
    Ld3: [
        list(ll[0] + (ll[1],))
        for ll in map(
            list, (product(combinations(ExecutionUnit.LOAD(), 2), ExecutionUnit.V()))
        )
    ],  # guessed
    (Ldr_Q): ExecutionUnit.LOAD(),  # for LDR (unsigned offset, Q)
    (Str_Q): ExecutionUnit.STORE(),  # for STR (signed offset, Q)
    (q_ldr1_stack, Q_Ld2_Lane_Post_Inc): list(
        map(list, product(ExecutionUnit.V(), ExecutionUnit.LOAD()))
    ),  # guessed
    Mov_xtov_d: ExecutionUnit.LOAD(),  # guessed
    # guessed
    d_stp_stack_with_inc: list(map(list, combinations(ExecutionUnit.STORE(), 2))),
    d_str_stack_with_inc: ExecutionUnit.STORE(),  # for STR (signed offset, D),
    b_ldr_stack_with_inc: ExecutionUnit.LOAD(),  # for LDR (unsigned offset, S)
    d_ldr_stack_with_inc: ExecutionUnit.LOAD(),  # for LDR (unsigned offset, D)
    is_qform_form_of(vmov): [],  # TODO: Can this be empty?
    is_dform_form_of(vmov): ExecutionUnit.V(),
    (vzip1, vzip2, vuzp1, vuzp2): ExecutionUnit.V(),
    # Arithmetic
    (add, add_imm): ExecutionUnit.I(),
    (add_shifted, add2): list(map(list, combinations(ExecutionUnit.I(), 2))),
    (umull_wform, mul_wform): ExecutionUnit.M(),
    (umaddl_wform): ExecutionUnit.M(),
    (
        lsr,
        bic,
        bic_reg,
        ubfx,
        add_sp_imm,
        and_imm,
        movk_imm,
        sub,
        mov,  # mov is guessed
        asr_wform,
        and_imm_wform,
        lsr_wform,
        eor_wform,
    ): ExecutionUnit.I(),
    (bfi): ExecutionUnit.SCALAR_I2,
    (nop): [],
    (tst_wform, subs_wform): ExecutionUnit.I(),
    # Load/Store
    (Ldr_X, x_ldr_stack_imm, ldr_sxtw_wform, ldr_const): ExecutionUnit.LOAD(),
    (Str_X, x_str_sp_imm): ExecutionUnit.STORE(),
    (x_stp_with_imm_sp, w_stp_with_imm_sp): ExecutionUnit.STORE(),
}

inverse_throughput = {
    # Neon Arithmetic
    (
        vmul,
        vmul_lane,
        vqrdmulh,
        vqrdmulh_lane,
        vmla,
        vmla_lane,
        vmls,
        vmls_lane,
        vqdmulh_lane,
        vumull,
        vumlal,
        vusra,
        vand,
        vbic,
        ASimdCompare,
        VShiftImmediateBasic,
        VShiftImmediateRounding,
    ): 1,
    (vadd, vsub, trn1, trn2): 1,
    Vins: 2,  # guessed
    (umov_d, mov_d): 2,  # guessed
    (mov_d01, mov_b00): 1,  # guessed
    fcsel: 1,
    # Neon Load/Store
    (Ldr_Q, Str_Q): 1,
    (q_ldr1_stack, Q_Ld2_Lane_Post_Inc): 3,  # guessed
    St4: 5,  # guessed
    Ld4: 5,  # guessed
    St3: 4,  # guessed
    Ld3: 4,  # guessed
    Mov_xtov_d: 1,  # guessed
    d_stp_stack_with_inc: 1,  # guessed
    d_str_stack_with_inc: 1,
    b_ldr_stack_with_inc: 1,  # for LDR (unsigned offset, S)
    d_ldr_stack_with_inc: 1,  # for LDR (unsigned offset, S)
    is_qform_form_of(vmov): 1,  # guessed
    is_dform_form_of(vmov): 1,
    (vzip1, vzip2, vuzp1, vuzp2): 1,
    # Arithmetic
    (add, add_imm): 1,
    (add_shifted, add2): 1,
    (umull_wform, mul_wform): 1,
    (umaddl_wform): 1,
    (
        lsr,
        bic,
        bic_reg,
        ubfx,
        add_sp_imm,
        and_imm,
        movk_imm,
        sub,
        mov,  # mov guessed
        asr_wform,
        and_imm_wform,
        lsr_wform,
        eor_wform,
    ): 1,
    (bfi): 1,
    (nop): 1,  # guessed
    (tst_wform, subs_wform): 1,
    # Load/Store
    (Ldr_X, x_ldr_stack_imm, ldr_sxtw_wform, ldr_const): 1,
    (Str_X, x_str_sp_imm): 1,
    (x_stp_with_imm_sp, w_stp_with_imm_sp): 1,
}

# REVISIT
default_latencies = {
    # Neon Arithmetic
    (
        vmul,
        vmul_lane,
        vqrdmulh,
        vqrdmulh_lane,
        vmls,
        vmls_lane,
        vmla,
        vmla_lane,
        vqdmulh_lane,
        vumull,
        vumlal,
        vusra,
    ): 3,
    VShiftImmediateRounding: 3,
    (VShiftImmediateBasic, vand, vbic, ASimdCompare): 2,
    (vadd, vsub, trn1, trn2): 2,
    Vins: 2,  # 2 or <= 9
    (umov_d, mov_d): 4,  # <= 7
    (mov_d01, mov_b00): 2,  # guessed
    fcsel: 2,
    # Neon Load/Store
    (Ldr_Q): 4,  # <= 7
    (Str_Q): 4,  # guessed
    St4: 4,  # guessed
    Ld4: 6,  # guessed
    St3: 3,  # guessed
    Ld3: 4,  # guessed
    (q_ldr1_stack, Q_Ld2_Lane_Post_Inc): 4,  # guessed
    Mov_xtov_d: 4,  # <=7, based on FMOV
    d_stp_stack_with_inc: 4,  # guessed
    d_str_stack_with_inc: 4,  # guessed
    b_ldr_stack_with_inc: 4,  # <=7, for LDR (unsigned offset, S)
    d_ldr_stack_with_inc: 4,  # <=7, for LDR (unsigned offset, D)
    is_qform_form_of(vmov): 1,
    is_dform_form_of(vmov): 2,
    (vzip1, vzip2, vuzp1, vuzp2): 2,
    # Arithmetic
    (add, add_imm): 1,
    (add_shifted, add2): 2,
    (umull_wform, mul_wform): 3,
    (umaddl_wform): 3,  # TODO: Add exception with 1 cycle
    (
        lsr,
        bic,
        bic_reg,
        ubfx,
        add_sp_imm,
        and_imm,
        movk_imm,
        sub,
        mov,  # mov guessed
        asr_wform,
        and_imm_wform,
        lsr_wform,
        eor_wform,
    ): 1,
    (bfi): 1,
    (nop): 1,  # guessed
    (tst_wform, subs_wform): 1,
    # Load/Store
    (Ldr_X, x_ldr_stack_imm): 3,  # <= 4, LDR (unsigned offset, 64-bit)
    (Str_X, x_str_sp_imm): 4,  # guessed
    (x_stp_with_imm_sp, w_stp_with_imm_sp): 4,  # guessed
    (ldr_const): 3,  # guessed
    (ldr_sxtw_wform): 3,  # <= 4
}


def get_latency(src, out_idx, dst):
    instclass_src = find_class(src)
    instclass_dst = find_class(dst)

    if (
        instclass_src == umaddl_wform
        and instclass_dst == umaddl_wform
        and src.args_out[0] == dst.args_in[2]
    ):
        return (
            3,
            lambda t_src, t_dst: t_dst.program_start_var == t_src.program_start_var + 1,
        )

    latency = lookup_multidict(default_latencies, src, instclass_src)
    return latency


def get_units(src):
    instclass_src = find_class(src)
    units = lookup_multidict(execution_units, src, instclass_src)
    if isinstance(units, list):
        return units
    else:
        return [units]


def get_inverse_throughput(src):
    instclass_src = find_class(src)
    return lookup_multidict(inverse_throughput, src, instclass_src)



#
# Copyright (c) 2022 Arm Limited
# Copyright (c) 2022 Hanno Becker
# Copyright (c) 2023 Amin Abdulrahman, Matthias Kannwischer
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Hanno Becker <hannobecker@posteo.de>
#

#
# Experimental model for high-end A-profile cores with 4 SIMD units
#

from enum import Enum
from slothy.helper import lookup_multidict
from slothy.targets.aarch64.aarch64_neon import (
    is_neon_instruction,
    find_class,
    aese_x4,
    aesr_x4,
    Ldp_X,
    Ldr_X,
    Str_X,
    Stp_X,
    Ldr_Q,
    Str_Q,
    Ldp_Q,
    Stp_Q,
    vuzp1,
    vuzp2,
    vzip1,
    Vrev,
    uaddlp,
    vmov,
    vmovi,
    vand,
    vadd,
    vxtn,
    veor3,
    VShiftImmediateBasic,
    vshl_d,
    vsli,
    vshrn,
    vusra,
    vmul,
    vumlal,
    vumull,
    vdup,
    AESInstruction,
    Transpose,
    aesr_x2,
    AArch64NeonLogical,
    AArch64BasicArithmetic,
    AArch64ConditionalSelect,
    AArch64ConditionalCompare,
    AArch64Logical,
    AArch64LogicalShifted,
    AArch64Move,
    AArch64Shift,
    Tst,
    AArch64ShiftedArithmetic,
    Fmov,
    fcsel,
    AArch64HighMultiply,
    AArch64Multiply,
    VecToGprMov,
)

issue_rate = 6


class ExecutionUnit(Enum):
    SCALAR_I0 = (0,)
    SCALAR_I1 = (1,)
    SCALAR_I2 = (2,)
    SCALAR_I3 = (3,)
    SCALAR_M0 = (2,)  # Overlaps with third I pipeline
    SCALAR_M1 = (3,)  # Overlaps with fourth I pipeline
    LSU0 = (4,)
    LSU1 = (5,)
    VEC0 = (6,)
    VEC1 = (7,)
    VEC2 = (8,)
    VEC3 = (9,)

    def __repr__(self):
        return self.name

    def I():  # noqa: E743
        return [
            ExecutionUnit.SCALAR_I0,
            ExecutionUnit.SCALAR_I1,
            ExecutionUnit.SCALAR_I2,
            ExecutionUnit.SCALAR_I3,
        ]

    def M():
        return [ExecutionUnit.SCALAR_M0, ExecutionUnit.SCALAR_M1]

    def V():
        return [
            ExecutionUnit.VEC0,
            ExecutionUnit.VEC1,
            ExecutionUnit.VEC2,
            ExecutionUnit.VEC3,
        ]

    def V0():
        return [ExecutionUnit.VEC0]

    def V1():
        return [ExecutionUnit.VEC1]

    def V13():
        return [ExecutionUnit.VEC1, ExecutionUnit.VEC3]

    def V01():
        return [ExecutionUnit.VEC0, ExecutionUnit.VEC1]

    def V02():
        return [ExecutionUnit.VEC0, ExecutionUnit.VEC2]

    def LSU():
        return [ExecutionUnit.LSU0, ExecutionUnit.LSU1]


# Opaque functions called by SLOTHY to add further microarchitecture-
# specific constraints which are not encapsulated by the general framework.
def add_further_constraints(slothy):
    if slothy.config.constraints.functional_only:
        return
    slothy.restrict_slots_for_instructions_by_property(
        is_neon_instruction, [0, 1, 2, 3]
    )
    slothy.restrict_slots_for_instructions_by_class([aesr_x4, aesr_x4], [0])


def has_min_max_objective(config):
    return False


def get_min_max_objective(slothy):
    return


# TODO: Copy-pasted from N1 model -- adjust


execution_units = {
    (Ldp_X, Ldr_X, Str_X, Stp_X, Ldr_Q, Str_Q, Ldp_Q, Stp_Q): ExecutionUnit.LSU(),
    (vuzp1, vuzp2, vzip1, Vrev, uaddlp): ExecutionUnit.V(),
    (vmov): ExecutionUnit.V(),
    VecToGprMov: ExecutionUnit.V(),
    (vmovi): ExecutionUnit.V(),
    (vand, vadd): ExecutionUnit.V(),
    (vxtn): ExecutionUnit.V(),
    veor3: ExecutionUnit.V(),
    fcsel: ExecutionUnit.V(),
    (
        VShiftImmediateBasic,
        vshl_d,
        vsli,
        vshrn,
    ): ExecutionUnit.V1(),  # TODO: Should be V13?
    vusra: ExecutionUnit.V1(),
    AESInstruction: ExecutionUnit.V(),
    Transpose: ExecutionUnit.V(),
    aesr_x2: ExecutionUnit.V(),
    aesr_x4: [ExecutionUnit.V()],  # Use all V-pipes
    aese_x4: [ExecutionUnit.V()],  # Use all V-pipes
    (vmul, vumlal, vumull): ExecutionUnit.V0(),
    AArch64NeonLogical: ExecutionUnit.V(),
    (
        AArch64BasicArithmetic,
        AArch64ConditionalSelect,
        AArch64ConditionalCompare,
        AArch64Logical,
        AArch64LogicalShifted,
        AArch64Move,
    ): ExecutionUnit.I(),
    AArch64Shift: ExecutionUnit.I(),
    Tst: ExecutionUnit.I(),
    AArch64ShiftedArithmetic: ExecutionUnit.M(),
    Fmov: ExecutionUnit.M(),
    (AArch64HighMultiply, AArch64Multiply): ExecutionUnit.M(),
    vdup: ExecutionUnit.M(),
}

inverse_throughput = {
    (Ldr_X, Str_X, Ldr_Q, Str_Q): 1,
    (Ldp_X, Stp_X, Ldp_Q, Stp_Q): 2,
    (vuzp1, vuzp2, vzip1, uaddlp, Vrev): 1,
    VecToGprMov: 1,
    Transpose: 1,
    veor3: 2,
    (vand, vadd): 1,
    (vmov): 1,
    AESInstruction: 1,
    aesr_x4: 1,
    aese_x4: 1,
    AArch64NeonLogical: 1,
    (vmovi): 1,
    (vxtn): 1,
    fcsel: 1,
    (VShiftImmediateBasic, vshl_d, vsli, vshrn): 1,
    (vmul): 2,
    vusra: 1,
    (vumlal, vumull): 1,
    (
        AArch64BasicArithmetic,
        AArch64ConditionalSelect,
        AArch64ConditionalCompare,
        AArch64Logical,
        AArch64LogicalShifted,
        AArch64Move,
    ): 1,
    AArch64Shift: 1,
    AArch64ShiftedArithmetic: 1,
    Tst: 1,
    Fmov: 1,
    (AArch64HighMultiply): 4,
    (AArch64Multiply): 3,
    (vdup): 1,
}

default_latencies = {
    (Ldp_X, Ldr_X, Ldr_Q, Ldp_Q): 4,
    (Stp_X, Str_X, Str_Q, Stp_Q): 2,
    (vuzp1, vuzp2, vzip1, Vrev, uaddlp): 2,
    VecToGprMov: 2,
    veor3: 2,
    (vxtn): 2,
    fcsel: 2,
    Transpose: 2,
    AESInstruction: 2,
    (aesr_x4, aese_x4): 2,
    AArch64NeonLogical: 2,
    (vand, vadd): 2,
    (vmov): 2,  # ???
    (vmovi): 2,
    (vmul): 5,
    vusra: 4,  # TODO: Add fwd path
    (vumlal, vumull): 4,  # TODO: Add fwd path
    (VShiftImmediateBasic, vshl_d, vsli, vshrn): 2,
    (
        AArch64BasicArithmetic,
        AArch64ConditionalSelect,
        AArch64ConditionalCompare,
        AArch64Logical,
        AArch64LogicalShifted,
        AArch64Move,
    ): 1,
    AArch64Shift: 1,
    AArch64ShiftedArithmetic: 2,
    Tst: 1,
    Fmov: 3,
    AArch64HighMultiply: 5,
    AArch64Multiply: 4,
    (vdup): 3,
}


def get_latency(src, out_idx, dst):
    instclass_src = find_class(src)
    latency = lookup_multidict(default_latencies, src, instclass_src)
    return latency


def get_units(src):
    instclass_src = find_class(src)
    units = lookup_multidict(execution_units, src, instclass_src)
    if isinstance(units, list):
        return units
    else:
        return [units]


def get_inverse_throughput(src):
    instclass_src = find_class(src)
    return lookup_multidict(inverse_throughput, src, instclass_src)



#
# Copyright (c) 2022 Arm Limited
# Copyright (c) 2022 Hanno Becker
# Copyright (c) 2023 Amin Abdulrahman, Matthias Kannwischer
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Hanno Becker <hannobecker@posteo.de>
#

"""
Experimental Cortex-A55 microarchitecture model for SLOTHY

Most data in this model is derived from the Cortex-A55 software optimization guide.
Some latency exceptions were manually identified through microbenchmarks.

.. warning::

    The data in this module is approximate and may contain errors.
"""
# ################################ NOTE ############################################ #
#                                                                                    #
#  WARNING: The data in this module is approximate and may contain errors.           #
#           They are _NOT_ an official software optimization guide for Cortex-A55.   #
#                                                                                    #
# ################################################################################## #

from enum import Enum
from slothy.helper import lookup_multidict
from slothy.targets.aarch64.aarch64_neon import (
    find_class,
    AArch64ConditionalCompare,
    Ldp_X,
    Ldr_X,
    Str_X,
    Stp_X,
    Ldr_Q,
    Ldr_D,
    Str_Q,
    vmov,
    vadd,
    vxtn,
    vshrn,
    vusra,
    vmul,
    Instruction,
    fcsel,
    Q_Ld2_Lane_Post_Inc,
    vmla,
    vmla_lane,
    vmls,
    vmls_lane,
    vmul_lane,
    vqrdmulh,
    vqrdmulh_lane,
    vqdmulh_lane,
    vqdmulh_vector,
    vbic,
    vbic_imm_shifted,
    q_ldr1_stack,
    q_ldr1_post_inc,
    Vmull,
    Vmlal,
    vushr,
    vsshr,
    vuxtl,
    vshl_d,
    vshl,
    VShiftImmediateRounding,
    St4,
    Ld4,
    St3,
    Ld3,
    St2,
    Ld2,
    umov_d,
    mov_d01,
    mov_b00,
    VecToGprMov,
    Mov_xtov_d,
    d_stp_stack_with_inc,
    d_str_stack_with_inc,
    b_ldr_stack_with_inc,
    d_ldr_stack_with_inc,
    is_qform_form_of,
    is_dform_form_of,
    trn1,
    trn2,
    vzip1,
    vzip2,
    vuzp1,
    vext,
    vuzp2,
    vsub,
    w_stp_with_imm_sp,
    ldr_const,
    ldr_sxtw_wform,
    umull_wform,
    mul_wform,
    umaddl_wform,
    lsr,
    bic,
    bic_reg,
    eor,
    eon,
    ror,
    eor_shifted,
    bic_shifted,
    bfi,
    add,
    add_imm,
    add_sp_imm,
    add2,
    add_shifted,
    adcs_to_zero,
    adcs_zero_r_to_zero,
    adcs_zero2,
    cmn,
    and_imm,
    nop,
    Vins,
    tst_wform,
    movk_imm,
    movz_imm,
    movz_imm_lsl,
    movw_imm,
    mov_imm,
    sub,
    sbcs_zero_to_zero,
    cmp_xzr2,
    cmp_imm,
    mov,
    ngc_zero,
    subs_wform,
    asr_wform,
    and_imm_wform,
    eor_wform,
    eon_wform,
    lsr_wform,
    ASimdCompare,
    and_twoarg,
    VShiftImmediateBasic,
    ubfx,
    AESInstruction,
    AArch64NeonCount,
    AArch64NeonLogical,
    AArch64NeonShiftInsert,
    vtbl,
    sub_imm,
    vuaddlv_sform,
    fmov_s_form,  # from double/single to gen reg
    cmp,
)

issue_rate = 2
llvm_mca_target = "cortex-a55"


class ExecutionUnit(Enum):
    """Enumeration of execution units in Cortex-A55 model"""

    SCALAR_ALU0 = 1
    SCALAR_ALU1 = 2
    SCALAR_MAC = 3
    SCALAR_LOAD = 4
    SCALAR_STORE = 5
    VEC0 = 6
    VEC1 = 7

    def __repr__(self):
        return self.name

    @classmethod
    def SCALAR(cls):
        """All scalar execution units"""
        return [ExecutionUnit.SCALAR_ALU0, ExecutionUnit.SCALAR_ALU1]

    @classmethod
    def SCALAR_MUL(cls):
        """All multiply-capable scalar execution units"""
        return [ExecutionUnit.SCALAR_MAC]


# Opaque function called by SLOTHY to add further microarchitecture-
# specific constraints which are not encapsulated by the general framework.
def add_further_constraints(slothy):
    if slothy.config.constraints.functional_only:
        return
    add_slot_constraints(slothy)
    add_st_hazard(slothy)


def add_slot_constraints(slothy):
    # Q-Form vector instructions are on slot 0 only
    slothy.restrict_slots_for_instructions_by_property(
        Instruction.is_q_form_vector_instruction, [0]
    )
    # fcsel and vld2 on slot 0 only
    slothy.restrict_slots_for_instructions_by_class([fcsel, Q_Ld2_Lane_Post_Inc], [0])


def add_st_hazard(slothy):
    def is_vec_st_st_pair(inst_a, inst_b):
        return inst_a.inst.is_vector_store() and inst_b.inst.is_vector_store()

    for t0, t1 in slothy.get_inst_pairs(cond=is_vec_st_st_pair):
        if t0.is_locked and t1.is_locked:
            continue
        slothy._Add(t0.cycle_start_var != t1.cycle_start_var + 1)


# Opaque function called by SLOTHY to add further microarchitecture-
# specific objectives.
def has_min_max_objective(config):
    """Adds Cortex-"""
    _ = config
    return False


def get_min_max_objective(slothy):
    _ = slothy
    return


execution_units = {
    # q-form vector instructions
    (
        vmla_lane,
        vmls_lane,
        vmul,
        vmul_lane,
        vqrdmulh,
        vqrdmulh_lane,
        vqdmulh_lane,
        vqdmulh_vector,
        Ldr_Q,
        Str_Q,
        q_ldr1_stack,
        q_ldr1_post_inc,
        Q_Ld2_Lane_Post_Inc,
        Vmull,
        Vmlal,
        vusra,
        vshrn,
        vxtn,
        vtbl,
        VShiftImmediateRounding,
        AArch64NeonLogical,
        vuaddlv_sform,
    ): [
        [ExecutionUnit.VEC0, ExecutionUnit.VEC1]
    ],  # these instructions use both VEC0 and VEC1
    St4: [
        [
            ExecutionUnit.VEC0,
            ExecutionUnit.VEC1,
            ExecutionUnit.SCALAR_LOAD,
            ExecutionUnit.SCALAR_STORE,
        ]
        + ExecutionUnit.SCALAR()
    ],
    Ld4: [
        [ExecutionUnit.VEC0, ExecutionUnit.VEC1, ExecutionUnit.SCALAR_LOAD]
        + ExecutionUnit.SCALAR()
    ],
    St3: [
        [
            ExecutionUnit.VEC0,
            ExecutionUnit.VEC1,
            ExecutionUnit.SCALAR_LOAD,
            ExecutionUnit.SCALAR_STORE,
        ]
        + ExecutionUnit.SCALAR()
    ],
    Ld3: [
        [ExecutionUnit.VEC0, ExecutionUnit.VEC1, ExecutionUnit.SCALAR_LOAD]
        + ExecutionUnit.SCALAR()
    ],
    St2: [
        [
            ExecutionUnit.VEC0,
            ExecutionUnit.VEC1,
            ExecutionUnit.SCALAR_LOAD,
            ExecutionUnit.SCALAR_STORE,
        ]
        + ExecutionUnit.SCALAR()
    ],
    Ld2: [
        [ExecutionUnit.VEC0, ExecutionUnit.VEC1, ExecutionUnit.SCALAR_LOAD]
        + ExecutionUnit.SCALAR()
    ],
    # non-q-form vector instructions
    (
        Ldr_D,
        umov_d,
        mov_d01,
        mov_b00,
        fcsel,
        VecToGprMov,
        Mov_xtov_d,
        d_stp_stack_with_inc,
        d_str_stack_with_inc,
        b_ldr_stack_with_inc,
        d_ldr_stack_with_inc,
        fmov_s_form,  # from double/single to gen reg
    ): [
        ExecutionUnit.VEC0,
        ExecutionUnit.VEC1,
    ],  # these instructions use VEC0 or VEC1
    is_qform_form_of(vmov): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vmov): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(trn1): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(trn1): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(trn2): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(trn2): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(ASimdCompare): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(ASimdCompare): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vzip1): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_qform_form_of(AArch64NeonCount): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(AArch64NeonCount): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_dform_form_of(vzip1): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vzip2): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vzip2): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vext): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vext): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vuzp1): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vuzp1): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vuzp2): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vuzp2): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vsub): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vsub): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vadd): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vadd): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vshl): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vshl): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vshrn): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vshrn): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vushr): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vushr): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vsshr): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vsshr): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vmla): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vmla): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    is_qform_form_of(vmls): [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_dform_form_of(vmls): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    vshl_d: [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    vuxtl: [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
    is_qform_form_of(AArch64NeonShiftInsert): [
        [ExecutionUnit.VEC0, ExecutionUnit.VEC1]
    ],
    is_dform_form_of(AArch64NeonShiftInsert): [ExecutionUnit.VEC0, ExecutionUnit.VEC1],
    (Stp_X, w_stp_with_imm_sp, Str_X): ExecutionUnit.SCALAR_STORE,
    (
        ldr_const,
        ldr_sxtw_wform,
        Ldr_X,
        Ldp_X,
    ): ExecutionUnit.SCALAR_LOAD,
    (umull_wform, mul_wform, umaddl_wform): ExecutionUnit.SCALAR_MUL(),
    (
        lsr,
        bic,
        bic_reg,
        bfi,
        ubfx,
        add,
        mov_imm,
        movw_imm,
        cmp_imm,
        eor,
        eon,
        eor_shifted,
        bic_shifted,
        ror,
        add_imm,
        add_sp_imm,
        add2,
        add_shifted,
        adcs_to_zero,
        adcs_zero_r_to_zero,
        adcs_zero2,
        cmn,
        and_imm,
        nop,
        Vins,
        tst_wform,
        movk_imm,
        movz_imm,
        movz_imm_lsl,
        sub,
        sub_imm,
        cmp,
        sbcs_zero_to_zero,
        cmp_xzr2,
        mov,
        ngc_zero,
        subs_wform,
        asr_wform,
        and_imm_wform,
        lsr_wform,
        eor_wform,
        eon_wform,
    ): ExecutionUnit.SCALAR(),
    AArch64ConditionalCompare: ExecutionUnit.SCALAR(),
    # NOTE: AESE/AESMC and AESD/AESIMC pairs can be dual-issued on A55 but this
    # is not modeled
    AESInstruction: [[ExecutionUnit.VEC0, ExecutionUnit.VEC1]],
}

inverse_throughput = {
    (
        vadd,
        vsub,
        vmov,
        vmul,
        vmul_lane,
        vqrdmulh,
        vqrdmulh_lane,
        vqdmulh_lane,
        vqdmulh_vector,
        Vmull,
        Vmlal,
        umov_d,
        vuaddlv_sform,
    ): 1,
    (sub_imm, cmp): 1,
    (
        vmla,
        vmla_lane,
        vmls,
        vmls_lane,
    ): 1,
    (vshl, vshl_d, vsshr, vushr, vuxtl): 1,
    (trn2, trn1, ASimdCompare): 1,
    (Ldr_D): 1,
    (Ldr_Q): 2,
    (AArch64NeonCount): 1,
    (Str_Q): 1,
    (tst_wform): 1,
    (nop, Vins, Ldr_X, Str_X): 1,
    Ldp_X: 2,
    St4: 5,
    St3: 3,
    St2: 2,
    Ld4: 9,
    Ld3: 6,
    Ld2: 4,
    vxtn: 1,
    vshrn: 2,
    vtbl: 1,  # N cycles (N = number of registers in the table)
    (fcsel): 1,
    (VecToGprMov, Mov_xtov_d): 1,
    (movk_imm, movz_imm, movz_imm_lsl, mov, mov_imm, movw_imm): 1,
    (d_stp_stack_with_inc, d_str_stack_with_inc): 1,
    (Stp_X, w_stp_with_imm_sp): 1,
    (ldr_const): 1,
    (ldr_sxtw_wform): 3,
    (lsr, lsr_wform, ror): 1,
    (umull_wform, mul_wform, umaddl_wform): 1,
    (and_twoarg, and_imm, and_imm_wform): 1,
    (
        add,
        add_imm,
        add2,
        add_shifted,
        add_sp_imm,
        adcs_to_zero,
        adcs_zero2,
        adcs_zero_r_to_zero,
        cmn,
    ): 1,
    (cmp_xzr2, cmp_imm, sub, subs_wform, asr_wform, sbcs_zero_to_zero, ngc_zero): 1,
    (bfi, ubfx): 1,
    VShiftImmediateRounding: 1,
    AArch64NeonShiftInsert: 1,
    (vusra): 1,
    AArch64NeonLogical: 1,
    vext: 1,
    (vuzp1, vuzp2): 1,
    (q_ldr1_stack, Q_Ld2_Lane_Post_Inc, q_ldr1_post_inc): 1,
    (b_ldr_stack_with_inc, d_ldr_stack_with_inc): 1,
    (mov_d01, mov_b00): 1,
    (vzip1, vzip2): 1,
    (eor_wform, eon_wform): 1,
    (eon, eor, bic, bic_reg, eor_shifted, bic_shifted): 1,
    AArch64ConditionalCompare: 1,
    AESInstruction: 1,
    fmov_s_form: 1,  # from double/single to gen reg
}

default_latencies = {
    vmov: 2,
    is_qform_form_of([vadd, vsub]): 3,
    is_dform_form_of([vadd, vsub]): 2,
    (trn1, trn2, ASimdCompare): 2,
    (
        vmul,
        vmul_lane,
        vqrdmulh,
        vqrdmulh_lane,
        vqdmulh_lane,
        vqdmulh_vector,
        Vmull,
        Vmlal,
    ): 4,
    (
        vmla,
        vmla_lane,
        vmls,
        vmls_lane,
    ): 4,
    (Ldr_D): 3,
    (Ldr_Q, Str_Q): 4,
    (sub_imm, cmp): 2,
    AArch64NeonCount: 2,
    St4: 5,
    St3: 3,
    St2: 2,
    # TODO: Add distinction between Q/D and B/H vs. D/S
    Ld2: 6,
    Ld3: 8,
    Ld4: 11,
    vxtn: 2,
    vshrn: 2,
    vtbl: 2,  # 2+N-1 cycles (N = number of registers in the table)
    (vuxtl): 2,
    (Str_X, Ldr_X): 4,
    Ldp_X: 4,
    (Vins, umov_d): 2,
    (tst_wform): 1,
    (fcsel): 2,
    (VecToGprMov, Mov_xtov_d): 2,
    (movk_imm, movz_imm, movz_imm_lsl, mov, mov_imm, movw_imm): 1,
    (d_stp_stack_with_inc, d_str_stack_with_inc): 1,
    (Stp_X, w_stp_with_imm_sp): 1,
    (ldr_const): 3,
    (ldr_sxtw_wform): 5,
    (lsr, lsr_wform): 1,
    (umull_wform, mul_wform, umaddl_wform): 3,
    (vuaddlv_sform): 3,
    (and_imm, and_imm_wform): 1,
    (add2, add_shifted, add_sp_imm): 2,
    (
        add,
        add_imm,
        adcs_to_zero,
        adcs_zero_r_to_zero,
        adcs_zero2,
        cmn,
        sub,
        subs_wform,
        asr_wform,
        sbcs_zero_to_zero,
        cmp_xzr2,
        ngc_zero,
        cmp_imm,
    ): 1,
    (bfi, ubfx): 2,
    VShiftImmediateRounding: 3,
    VShiftImmediateBasic: 2,
    AArch64NeonShiftInsert: 2,
    (vusra): 3,
    AArch64NeonLogical: 1,
    vext: 2,
    (vuzp1, vuzp2): 2,
    (q_ldr1_stack, Q_Ld2_Lane_Post_Inc, q_ldr1_post_inc): 3,
    (b_ldr_stack_with_inc, d_ldr_stack_with_inc): 3,
    (mov_d01, mov_b00): 2,
    (vzip1, vzip2): 2,
    (eor_wform, eon_wform): 1,
    # According to SWOG, this is 2 cycles, byt if the output is used as a
    # _non-shifted_ input to the next instruction, the effective latency
    # seems to be 1 cycle. See https://eprint.iacr.org/2022/1243.pdf
    (eor_shifted, bic_shifted): 1,
    (eon, ror, eor, bic, bic_reg): 1,
    AArch64ConditionalCompare: 1,
    # NOTE: AESE/AESMC and AESD/AESIMC pairs can be dual-issued on A55 but this
    # is not modeled
    AESInstruction: 2,
    fmov_s_form: 1,  # from double/single to gen reg
}


def get_latency(src, out_idx, dst):
    _ = out_idx  # out_idx unused

    instclass_src = find_class(src)
    instclass_dst = find_class(dst)

    latency = lookup_multidict(default_latencies, src, instclass_src)

    if instclass_dst in [trn1, trn2, vzip1, vzip2, vuzp1, vuzp2, fcsel] and latency < 3:
        latency += 1

    if [instclass_src, instclass_dst] in [
        [lsr, mul_wform],
        [lsr, umaddl_wform],
        [vbic, vusra],
        [vbic_imm_shifted, vusra],
    ]:
        latency += 1

    if (
        isinstance(src, Vmlal)
        and isinstance(dst, Vmlal)
        and src.args_in_out[0] == dst.args_in_out[0]
    ):
        return (
            4,
            lambda t_src, t_dst: t_dst.program_start_var == t_src.program_start_var + 2,
        )

    if (
        instclass_src == umaddl_wform
        and instclass_dst == umaddl_wform
        and src.args_out[0] == dst.args_out[0]
    ):
        return (
            3,
            lambda t_src, t_dst: t_dst.program_start_var == t_src.program_start_var + 1,
        )

    return latency


def get_units(src):
    instclass_src = find_class(src)
    units = lookup_multidict(execution_units, src, instclass_src)
    if isinstance(units, list):
        return units
    return [units]


def get_inverse_throughput(src):
    instclass_src = find_class(src)
    return lookup_multidict(inverse_throughput, src, instclass_src)



#
# Copyright (c) 2022 Arm Limited
# Copyright (c) 2022 Hanno Becker
# Copyright (c) 2023 Amin Abdulrahman, Matthias Kannwischer
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Hanno Becker <hannobecker@posteo.de>
#

"""
Convenience module for querying available architecture
and microarchitecture models for SLOTHY.
"""

from slothy.targets.arm_v7m import arch_v7m as Arch_Armv7M
from slothy.targets.arm_v7m import cortex_m7 as Target_CortexM7

from slothy.targets.arm_v81m import arch_v81m as Arch_Armv81M
from slothy.targets.arm_v81m import cortex_m55r1 as Target_CortexM55r1
from slothy.targets.arm_v81m import cortex_m85r1 as Target_CortexM85r1
from slothy.targets.arm_v81m import helium_experimental as Target_Helium_Experimental

from slothy.targets.aarch64 import aarch64_neon as AArch64_Neon
from slothy.targets.aarch64 import cortex_a55 as Target_CortexA55
from slothy.targets.aarch64 import cortex_a72_frontend as Target_CortexA72_Frontend
from slothy.targets.aarch64 import (
    neoverse_n1_experimental as Target_NeoverseN1_Experimental,
)
from slothy.targets.aarch64 import aarch64_big_experimental as Target_Big_Experimental
from slothy.targets.aarch64 import (
    apple_m1_firestorm_experimental as Target_AppleM1_Firestorm_Experimental,
)
from slothy.targets.aarch64 import (
    apple_m1_icestorm_experimental as Target_AppleM1_Icestorm_Experimental,
)


class UnknownTarget(Exception):
    """Exception raised when an unknown architecture or microarchitecture
    is requested."""


class Archery:
    """This is a small helper class for querying architectures"""

    _archs = {
        "Arm_v7M": Arch_Armv7M,
        "Arm_v81M": Arch_Armv81M,
        "Arm_AArch64": AArch64_Neon,
    }

    _targets = {
        "Arm_Cortex_M55": Target_CortexM55r1,
        "Arm_Cortex_M85": Target_CortexM85r1,
        "Arm_Helium_Experimental": Target_Helium_Experimental,
        "Arm_Cortex_M7": Target_CortexM7,
        "Arm_Cortex_A55": Target_CortexA55,
        "Arm_Cortex_A72_frontend": Target_CortexA72_Frontend,
        "Arm_Neoverse_N1_experimental": Target_NeoverseN1_Experimental,
        "Arm_Big_experimental": Target_Big_Experimental,
        "Apple_M1_firestorm_experimental": Target_AppleM1_Firestorm_Experimental,
        "Apple_M1_icestorm_experimental": Target_AppleM1_Icestorm_Experimental,
    }

    @staticmethod
    def list_archs():
        """Lists all available architectures"""
        return list(Archery._archs.keys())

    @staticmethod
    def list_targets():
        """Lists all available targets"""
        return list(Archery._targets.keys())

    @staticmethod
    def get_arch(name):
        """Query an architecture by name"""
        arch = Archery._archs.get(name, None)
        if arch is None:
            raise UnknownTarget(
                f"Could not find architecture {name}. "
                f"Known architectures are {list(Archery._archs.keys())}"
            )
        return arch

    @staticmethod
    def get_target(name):
        """Query a target by name"""
        target = Archery._targets.get(name, None)
        if target is None:
            raise UnknownTarget(
                f"Could not find target {name}. "
                f"Known targets are {list(Archery._targets.keys())}"
            )
        return target



#!/usr/bin/env python3

# Copyright (c) 2022 Arm Limited
# Copyright (c) 2022 Hanno Becker
# SPDX-License-Identifier: MIT

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import argparse
import logging
import time
import os

from slothy import Slothy, Archery


class CmdLineException(Exception):
    """Exception thrown when a problem is encountered with the command line parameters"""


def validate_output_path(output_path):
    """Validate that the output path can be written to."""
    if output_path is None:
        return
    output_dir = os.path.dirname(output_path) or "."
    if not os.path.exists(output_dir):
        raise CmdLineException(f"Output directory '{output_dir}' does not exist")
    if not os.path.isdir(output_dir):
        raise CmdLineException(f"'{output_dir}' is not a directory")


def main():
    """Main entry point for the slothy-cli command."""

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "arch", type=str, choices=Archery.list_archs(), help="The target architecture"
    )
    parser.add_argument(
        "target",
        type=str,
        choices=Archery.list_targets(),
        help="The target microarchitecture",
    )
    parser.add_argument("input", type=str, help="The name of the assembly source file.")
    parser.add_argument(
        "-d", "--debug", default=False, action="store_true", help="Show debug output"
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        help="The name of the file to write the generated assembly to. "
        "If unspecified, the assembly will be printed on the standard output.",
    )
    parser.add_argument(
        "-c",
        "--config",
        default=[],
        action="append",
        nargs="*",
        metavar="OPTION=VALUE",
        help="""Set SLOTHY configuration value (can be used multiple times).
                See Python API documentation for details here:
                https://slothy-optimizer.github.io/slothy/apidocs/slothy/slothy.core.config.html""",  # noqa: E501
    )
    parser.add_argument(
        "-l",
        "--loop",
        default=[],
        action="append",
        type=str,
        help="""The starting label for the loop to optimize. This is mutually
                exclusive with -s/--start and -e/--end, which allowv you to specify
                the code to optimize via start/end separately.""",
    )
    parser.add_argument("--fusion", default=False, action="store_true")
    parser.add_argument("--fusion-only", default=False, action="store_true")
    parser.add_argument(
        "--unfold",
        default=False,
        action="store_true",
        help="""Unfold macros and/or register aliases, but don't optimize.
                                See also --unfold-macros and --unfold-aliases.""",
    )
    parser.add_argument(
        "--unfold-macros",
        type=bool,
        default=True,
        help="""If --unfold is set, unfold assembly macros.""",
    )
    parser.add_argument(
        "--unfold-aliases",
        type=bool,
        default=False,
        help="""If --unfold is set, unfold register aliases.""",
    )
    parser.add_argument(
        "-s",
        "--start",
        default=None,
        type=str,
        help="""The label or line at which the to code to optimize begins.
                This is mutually exclusive with -l/--loop.""",
    )
    parser.add_argument(
        "-e",
        "--end",
        default=None,
        type=str,
        help="""The label or line at which the to code to optimize ends
                This is mutually exclusive with -l/--loop.""",
    )
    parser.add_argument(
        "--se",
        default=[],
        action="append",
        nargs="*",
        metavar="START,END",
        help="""Specify start,end label pairs. Can be provided multiple times.
                Mutually exclusive with -l/--loop, -s/--start, and -e/--end.""",
    )
    parser.add_argument(
        "-r",
        "--rename-function",
        default=None,
        type=str,
        help="""Perform function renaming. Format: 'old_func_name,new_func_name'""",
    )
    parser.add_argument(
        "--silent",
        default=False,
        action="store_true",
        help="""Silent mode: Only print warnings and errors""",
    )
    parser.add_argument(
        "--log",
        default=False,
        action="store_true",
        help="""Write logging output to file""",
    )
    parser.add_argument(
        "--logdir", default=".", type=str, help="""Directory to store log output to"""
    )
    parser.add_argument(
        "--logfile",
        default=None,
        type=str,
        help="""File to write logging output to. Can be omitted, "\
                "in which case a generic name with timestamp is used""",
    )

    args = parser.parse_args()

    if args.se and (args.loop or args.start or args.end):
        raise CmdLineException(
            "--se cannot be used together with -l/--loop, -s/--start, or -e/--end"
        )

    handlers = []

    h_err = logging.StreamHandler(sys.stderr)
    h_err.setLevel(logging.WARNING)
    handlers.append(h_err)

    if args.log is True and args.logfile is None:
        # By default, use time stamp and input file
        if args.output is not None:
            file_base = os.path.basename(args.output).replace(".", "_")
        else:
            file_base = os.path.basename(args.input).replace(".", "_")
        args.logfile = f"slothy_log_{int(time.time())}_{file_base}.log"
        logfile = f"{args.logdir}/{args.logfile}"

    if args.silent is False:
        h_info = logging.StreamHandler(sys.stdout)
        h_info.setLevel(logging.DEBUG)
        h_info.addFilter(lambda r: r.levelno == logging.INFO)
        handlers.append(h_info)

    if args.debug:
        h_verbose = logging.StreamHandler(sys.stdout)
        h_verbose.setLevel(logging.DEBUG)
        h_verbose.addFilter(lambda r: r.levelno < logging.INFO)
        handlers.append(h_verbose)

    if args.log:
        h_log = logging.FileHandler(logfile)
        h_log.setLevel(logging.DEBUG)
        handlers.append(h_log)

    if args.debug:
        base_level = logging.DEBUG
    else:
        base_level = logging.INFO

    logging.basicConfig(
        level=base_level,
        handlers=handlers,
    )

    logger = logging.getLogger("slothy-cli")

    arch = Archery.get_arch(args.arch)
    target = Archery.get_target(args.target)
    slothy = Slothy(arch, target, logger=logger)

    def parse_config_value_as(val, ty):
        def parse_as_float(val):
            try:
                res = float(val)
                return res
            except ValueError:
                return None

        def check_ty(ty_real):
            if ty is None or ty is type(None) or ty == ty_real:
                return
            raise CmdLineException(
                f"Configuration value {val} isn't correctly typed -- "
                f"expected {ty}, but got {ty_real}"
            )

        if val == "":
            raise CmdLineException("Invalid configuration value")
        logger.debug("Parsing configuration value %s with expected type %s", val, ty)
        if val.isdigit():
            check_ty(int)
            logger.debug("Value %s parsed as integer", val)
            return int(val)
        if val.lower() == "true":
            check_ty(bool)
            logger.debug("Value %s parsed as Boolean", val)
            return True
        if val.lower() == "false":
            check_ty(bool)
            logger.debug("Value %s parsed as Boolean", val)
            return False
        # Try to parse as RegisterType
        ty = arch.RegisterType.from_string(val)
        if ty is not None:
            logger.debug("Value %s parsed as RegisterType", val)
            return ty
        f = parse_as_float(val)
        if f is not None:
            check_ty(float)
            logger.debug("Value %s parsed as float", val)
            return f
        if val[0] == "[" and val[-1] == "]":
            check_ty(list)
            val = val[1:-1].split(",")
            val = list(map(str.strip, val))

            # Find numeric suffix (e.g. x30 -> ('x', 30))
            def split_numeric_suffix(v):
                # Find first digit
                i = next((i for (i, c) in enumerate(v) if c.isdigit()), len(v))
                return v[:i], v[i:]

            # Check for range entries (e.g. 'x10--x18')
            def unfold_range(v):
                if "--" not in v:
                    return [v]
                vs = v.split("--")
                if not len(vs) == 2:
                    logger.debug("Invalid range entry %s -- ignore", v)
                    return [v]
                # Find numeric suffix
                v0, v1 = vs
                v0, v0i = split_numeric_suffix(v0)
                v1, v1i = split_numeric_suffix(v1)
                if v0 != v1:
                    raise CmdLineException(f"Invalid range expression {v}")
                # Ranges are inclusive
                res = [f"{v0}{i}" for i in range(int(v0i), int(v1i) + 1)]
                logger.debug("Decoded range entry %s to %s", v, res)
                return res

            val = [r for v in val for r in unfold_range(v)]
            logger.debug("Parsing %s is a list -- parse recursively", val)
            return [parse_config_value_as(v, None) for v in val]
        if val[0] == "{" and val[-1] == "}":
            check_ty(dict)
            kvs = val[1:-1].split(",")
            kvs = [kv.split(":") for kv in kvs]
            for kv in kvs:
                if not len(kv) == 2:
                    raise CmdLineException("Invalid dictionary entry")
            logger.debug("Parsing %s is a dictionary -- parse recursively", val)
            return {
                parse_config_value_as(k, None): parse_config_value_as(v, None)
                for k, v in kvs
            }
        logger.debug("Parsing %s as string", val)
        return val

    # A plain '-c' without arguments should list all available configuration options
    if [] in args.config:
        slothy.config.list_options()
        return

    if args.rename_function:
        args.rename_function = args.rename_function.split(",")
        if len(args.rename_function) != 2:
            logger.error("Invalid function renaming argument")
            return

    # Parse and set Slothy configuration

    def setattr_recursive(obj, attr, val):
        attr.strip()
        # If attr starts with
        attrs = attr.split(".")
        while len(attrs) > 1:
            obj = getattr(obj, attrs.pop(0))
        attr = attrs.pop(0)
        val = parse_config_value_as(val, type(getattr(obj, attr)))
        logger.info("Setting configuration option %s to value %s", attr, val)
        setattr(obj, attr, val)

    def check_list_of_fixed_len_list(lst):
        invalid = next(filter(lambda o: len(o) != 1, lst), None)
        if invalid is not None:
            raise CmdLineException(f"Invalid configuration argument {invalid} in {lst}")

    check_list_of_fixed_len_list(args.config)
    config_kv_pairs = [c[0].split("=") for c in args.config]
    for kv in config_kv_pairs:
        # We allow shorthands for boolean configurations
        # "-c config.options" as a shorthand for "-c config.options=True"
        # "-c /config.options" as a shorthand for "-c config.options=False"
        # '!' would be more intuitive, but this confuses some shells.
        negate_char = "/"
        if len(kv) == 1:
            kv[0] = kv[0].strip()
            if kv[0][0] == negate_char:
                val = False
                kv[0] = kv[0][1:]
            else:
                val = True
            setattr_recursive(slothy.config, kv[0], str(val))
        elif len(kv) == 2:
            setattr_recursive(slothy.config, kv[0], kv[1])
        else:
            raise CmdLineException(f"Invalid configuration {kv}")

    # Validate output path before starting optimization
    validate_output_path(args.output)

    # Read input
    slothy.load_source_from_file(args.input)

    done = False

    # Unfold only?
    if args.unfold is True:
        slothy.unfold(
            start=args.start,
            end=args.end,
            macros=args.unfold_macros,
            aliases=args.unfold_aliases,
        )
        done = True

    # Fusion
    if done is False and args.fusion is True:
        if len(args.loop) > 0:
            for lll in args.loop:
                slothy.fusion_loop(lll)
        if args.fusion_only:
            done = True

    # Optimize
    if done is False:
        if len(args.loop) > 0:
            for lll in args.loop:
                slothy.optimize_loop(lll)
        elif args.se:
            for pair in args.se:
                parts = pair.split(",", 1)
                if len(parts) != 2:
                    raise CmdLineException(
                        f"Invalid --se pair '{pair}', expected start,end"
                    )
                slothy.optimize(start=parts[0], end=parts[1])
        else:
            slothy.optimize(start=args.start, end=args.end)

    # Rename
    if args.rename_function:
        slothy.rename_function(args.rename_function[0], args.rename_function[1])
        slothy.rename_function(
            "_" + args.rename_function[0], "_" + args.rename_function[1]
        )

    # Write output
    if args.output is not None:
        slothy.write_source_to_file(args.output)
    else:
        print(slothy.get_source_as_string())


if __name__ == "__main__":
    main()



from slothy.core.slothy import Slothy
from slothy.core.core import SlothyException
from slothy.core.config import Config
from slothy.targets.query import Archery


__all__ = ["Slothy", "SlothyException", "Config", "Archery"]
